import { SolAmount } from './Amount';
import type { Context } from './Context';
import type { Instruction, WrappedInstruction } from './Instruction';
import type { RpcConfirmTransactionOptions, RpcConfirmTransactionResult, RpcSendTransactionOptions } from './RpcInterface';
import { Signer } from './Signer';
import { AddressLookupTableInput, Blockhash, BlockhashWithExpiryBlockHeight, Transaction, TransactionSignature, TransactionVersion } from './Transaction';
export type TransactionBuilderItemsInput = WrappedInstruction | WrappedInstruction[] | TransactionBuilder | TransactionBuilder[];
export type TransactionBuilderOptions = {
    version?: TransactionVersion;
    addressLookupTables?: AddressLookupTableInput[];
    blockhash?: Blockhash | BlockhashWithExpiryBlockHeight;
};
export type TransactionBuilderSendAndConfirmOptions = {
    send?: RpcSendTransactionOptions;
    confirm?: Partial<RpcConfirmTransactionOptions>;
};
export declare class TransactionBuilder {
    readonly context: Pick<Context, 'rpc' | 'transactions' | 'payer'>;
    readonly items: WrappedInstruction[];
    readonly options: TransactionBuilderOptions;
    constructor(context: Pick<Context, 'rpc' | 'transactions' | 'payer'>, items?: WrappedInstruction[], options?: TransactionBuilderOptions);
    empty(): TransactionBuilder;
    setItems(input: TransactionBuilderItemsInput): TransactionBuilder;
    prepend(input: TransactionBuilderItemsInput): TransactionBuilder;
    append(input: TransactionBuilderItemsInput): TransactionBuilder;
    add(input: TransactionBuilderItemsInput): TransactionBuilder;
    splitByIndex(index: number): [TransactionBuilder, TransactionBuilder];
    /**
     * Split the builder into multiple builders, such that
     * each of them should fit in a single transaction.
     *
     * This method is unsafe for several reasons:
     * - Because transactions are atomic, splitting the builder
     *   into multiple transactions may cause undesired side effects.
     *   For example, if the first transaction succeeds but the second
     *   one fails, you may end up with an inconsistent account state.
     *   This is why it is recommended to manually split your transactions
     *   such that each of them is valid on its own.
     * - It can only split the instructions of the builder. Meaning that,
     *   if the builder has a single instruction that is too big to fit in
     *   a single transaction, it will not be able to split it.
     */
    unsafeSplitByTransactionSize(): TransactionBuilder[];
    setVersion(version: TransactionVersion): TransactionBuilder;
    useLegacyVersion(): TransactionBuilder;
    useV0(): TransactionBuilder;
    setAddressLookupTables(addressLookupTables: AddressLookupTableInput[]): TransactionBuilder;
    getBlockhash(): Blockhash | undefined;
    setBlockhash(blockhash: Blockhash | BlockhashWithExpiryBlockHeight): TransactionBuilder;
    setLatestBlockhash(): Promise<TransactionBuilder>;
    getInstructions(): Instruction[];
    getSigners(): Signer[];
    getBytesCreatedOnChain(): number;
    getRentCreatedOnChain(): Promise<SolAmount>;
    getTransactionSize(): number;
    minimumTransactionsRequired(): number;
    fitsInOneTransaction(): boolean;
    build(): Transaction;
    buildWithLatestBlockhash(): Promise<Transaction>;
    buildAndSign(): Promise<Transaction>;
    send(options?: RpcSendTransactionOptions): Promise<TransactionSignature>;
    sendAndConfirm(options?: TransactionBuilderSendAndConfirmOptions): Promise<{
        signature: TransactionSignature;
        result: RpcConfirmTransactionResult;
    }>;
    protected parseItems(input: TransactionBuilderItemsInput): WrappedInstruction[];
}
export declare const transactionBuilder: (context: Pick<Context, 'rpc' | 'transactions' | 'payer'>, items?: WrappedInstruction[]) => TransactionBuilder;
